from aiohttp import ClientSession
import asyncio
from datetime import datetime
from enum import Enum
from .exceptions import TreillageHTTPException, TreillageException
import hashlib
import jwt


class TokenRequestType(Enum):
    KEY = 'key'
    SESSION = 'session'


class TokenManager:

    def __init__(self, credentials, base_url):
        self.__API_KEY = credentials.key
        self.__API_SECRET = credentials.secret
        self.__auth_url = '/'.join([base_url, "session"])
        self.__access_token = None
        self.__access_token_expiry = None
        self.__refresh_token = None
        self.__refresh_token_expiry = None
        self.__refresh_token_ttl = None
        self.__user_id = None
        self.__org_id = None

    @classmethod
    async def create(cls, credentials, base_url):
        self = TokenManager(credentials, base_url)
        self.__set_tokens(
            await self.__request_tokens(request_type=TokenRequestType.KEY)
        )
        return self

    @property
    def access_token(self) -> str:
        return self.__access_token

    @property
    def access_token_expiry(self) -> int:
        return self.__access_token_expiry

    @property
    def refresh_token(self) -> str:
        return self.__refresh_token

    @staticmethod
    def get_timestamp():
        timestamp = datetime.utcnow().isoformat()
        return timestamp[:-3] + 'Z'

    @staticmethod
    def create_hash(api_key, timestamp, api_secret):
        hash_string = '/'.join([api_key, timestamp, api_secret])
        return hashlib.md5(hash_string.encode()).hexdigest()

    async def __request_tokens(self, request_type: TokenRequestType):
        timestamp = self.get_timestamp()
        api_hash = self.create_hash(self.__API_KEY,
                                    timestamp,
                                    self.__API_SECRET)

        if request_type == TokenRequestType.KEY:
            request_body = {
                'mode': request_type.value,
                'apiKey': self.__API_KEY,
                'apiHash': api_hash,
                'apiTimestamp': timestamp,
            }
        elif request_type == TokenRequestType.SESSION:
            request_body = {
                'mode': request_type.value,
                'apiKey': self.__API_KEY,
                'apiHash': api_hash,
                'apiTimestamp': timestamp,
                'userId': self.__user_id,
                'orgId': self.__org_id,
                'sessionId': self.__refresh_token
            }
        else:
            raise TreillageException(msg="Invalid Token Request Type")
        try:
            async with ClientSession() as session:
                async with session.post(self.__auth_url,
                                        json=request_body) as resp:
                    if resp.status == 200:
                        return await resp.json()
                    else:
                        raise TreillageHTTPException(
                            resp.status,
                            msg='Failed to get API tokens',
                            url=self.__auth_url
                        )
        finally:
            await asyncio.sleep(0.250)

    def __set_tokens(self, tokens: dict):
        self.__access_token = tokens["accessToken"]
        self.__access_token_expiry = jwt.decode(jwt=tokens['accessToken'],
                                                algorithms=["RS256"],
                                                options={
                                                    "verify_signature": False,
                                                    "verify_aud": False,
                                                    "verify_iss": False
                                                }
                                                )['exp']
        self.__refresh_token = tokens["refreshToken"]
        self.__refresh_token_expiry = tokens["refreshTokenExpiry"]
        self.__refresh_token_ttl = tokens["refreshTokenTtl"]
        self.__user_id = tokens["userId"]
        self.__org_id = tokens["orgId"]

    async def __aenter__(self):
        self.__set_tokens(await self.__request_tokens(TokenRequestType.KEY))
        return self

    async def __aexit__(self, exception_type, exception_value, traceback):
        pass

    async def refresh_access_token(self):
        self.__set_tokens(
            await self.__request_tokens(TokenRequestType.SESSION)
        )



from .exceptions import TreillageException
import yaml


class Credential:
    def __init__(self, key, secret):
        self.key = key
        self.secret = secret

    @classmethod
    def get_credentials(cls, credentials_file):
        with open(credentials_file) as file:
            credentials_data = yaml.safe_load(file)
        if 'key' not in credentials_data or 'secret' not in credentials_data:
            raise TreillageException(
                msg=("Credentials file must contain 'key' and 'secret'"
                     "as top level key-value pairs"))
        else:
            return Credential(
                credentials_data['key'],
                credentials_data['secret']
            )



import asyncio
from math import log2, ceil
import random
import time


class RateLimiter:
    def __init__(self,
                 token_rate: int = 8,
                 max_backoff_time: int = 64):
        self.__tokens = token_rate
        self.__MAX_TOKENS = token_rate
        self.__token_rate = token_rate
        self.__last_update = time.monotonic()
        self.__max_backoff_time = max_backoff_time
        self.__last_try_success = True
        self.__failed_attempts = 0

    async def get_token(self):
        # backoff if a rate limit error was received
        if self.__failed_attempts > 0:
            await asyncio.sleep(
                self.__get_backoff_time_ms() / 1000  # convert ms to seconds
            )
        # Regen tokens if there aren't any available
        while self.__tokens < 1:
            await self.__wait_for_token()
        # Consume a token
        self.__tokens -= 1

    async def __wait_for_token(self):
        # Check if there are any tokens available to regen
        if not self.__add_new_token():
            # If not wait the amount of time it take to regen one token
            await asyncio.sleep(1 / self.__token_rate)

    @property
    def tokens(self) -> int:
        return self.__tokens

    @tokens.setter
    def tokens(self, i):
        # Must be between 0 and MAX_TOKENS
        self.__tokens = max(min(i, self.__MAX_TOKENS), 0)

    def last_try_success(self, was_success: bool):
        self.__last_try_success = was_success
        if not was_success:
            self.__failed_attempts += 1
        else:
            self.__failed_attempts = max(
                0,
                self.__failed_attempts - self.__MAX_TOKENS / 3
            )

    def __get_backoff_time_ms(self) -> float:
        """
        Return time in milliseconds to backoff after a failed request

        Returns a random amount of time in milliseconds between 100ms and an
        upper limit of 2^num_failed_attempts * 100. To avoid overflow, the
        maximum exponent is limited so that max_backoff_time is never exceeded
        """
        return random.randint(
            100,  # minimum 100ms
            # The ceil function is working in tenths of seconds,
            # so max_backoff_time must be converted from seconds to tenths of
            # seconds and the result must be converted from tenths to
            # milliseconds (thousandths of seconds).
            ceil(2 ** (
                min(
                    log2(self.__max_backoff_time * 10),
                    self.__failed_attempts
                )
            )) * 100
        )

    def __add_new_token(self) -> bool:
        now = time.monotonic()
        time_since_update = now - self.__last_update
        new_tokens = time_since_update * self.__token_rate
        if self.__tokens + new_tokens >= 1:
            self.tokens += new_tokens
            self.__last_update = now
            return True
        else:
            return False



from .. import ConnectionManager, TreillageRateLimitException


async def list_paginator(
        connection: ConnectionManager,
        endpoint: str,
        params: dict
):
    has_more = True
    params['offset'] = 0
    params['limit'] = 100

    while has_more:
        try:
            resp = await connection.get(endpoint, params)
            has_more = resp['hasMore']
            params['offset'] += params['limit']
            for item in resp['items']:
                yield item
        except TreillageRateLimitException:
            pass



from ._version import get_versions
from .treillage import Treillage
from .treillage import BaseURL
from .exceptions import *
from .credential import Credential
from .ratelimiter import RateLimiter
from .token_manager import TokenManager
from .connection_manager import ConnectionManager
from .connection_manager import retry_on_rate_limit

__version__ = get_versions()['version']
del get_versions



from typing import List
from .. import ConnectionManager
from .list_paginator import list_paginator


# * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
#                              Documents
# * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

async def get_document(connection: ConnectionManager,
                       document_id: str,
                       requested_fields: List[str] = None):
    endpoint = f"/core/documents/{document_id}"
    params = dict()
    if requested_fields:
        fields = ','.join(*[requested_fields])
        params['requestedFields'] = fields

    return await connection.get(endpoint, params)


async def get_document_list(connection: ConnectionManager,
                            requested_fields: List[str] = None,
                            folder_id: str = None):
    endpoint = "/core/documents/"

    params = dict()
    if requested_fields:
        fields = ','.join(*[requested_fields])
        params['requestedFields'] = fields
    if folder_id:
        params['folderId'] = folder_id

    async for document in list_paginator(connection, endpoint, params):
        yield document


async def delete_document(connection: ConnectionManager,
                          document_id: str):
    endpoint = f"/core/documents/{document_id}"
    await connection.delete(endpoint)



from .credential import Credential
from .connection_manager import ConnectionManager
from enum import Enum
from typing import Union


class BaseURL(Enum):
    UNITED_STATES = "https://api.filevine.io"
    CANADA = "https://api.filevine.ca"


class Treillage:
    def __init__(self,
                 credentials_file: str,
                 base_url: Union[str, BaseURL] = BaseURL.UNITED_STATES.value,
                 # Number of parallel connections to each host:port endpoint
                 max_connections: int = None,
                 # Maximum requests per second allowed by the rate-limiter
                 requests_per_second: int = None):
        self.__credential = Credential.get_credentials(credentials_file)
        if isinstance(base_url, BaseURL):
            self.__base_url = base_url.value
        elif isinstance(base_url, str):
            self.__base_url = base_url
        self.__max_connections = max_connections
        self.__requests_per_second = requests_per_second
        self.__conn = None

    @property
    def conn(self) -> ConnectionManager:
        return self.__conn

    async def __async_init(self):
        self.__conn = await ConnectionManager.create(
            self.__base_url,
            self.__credential,
            self.__max_connections,
            self.__requests_per_second
        )

    @classmethod
    async def create(
            cls,
            credentials_file: str,
            base_url: Union[str, BaseURL] = BaseURL.UNITED_STATES.value,
            max_connections: int = None,
            requests_per_second: int = None,
    ):
        self = Treillage(credentials_file,
                         base_url,
                         max_connections,
                         requests_per_second)
        await self.__async_init()
        return self

    async def close(self):
        await self.__conn.close()

    async def __aenter__(self):
        await self.__async_init()
        return self

    async def __aexit__(self, exception_type, exception_value, traceback):
        await self.close()
